package lab.game;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import javafx.scene.canvas.GraphicsContext;
import lab.Setting;
import lab.game.Ufo.UfoDestroyLog;
import lombok.extern.log4j.Log4j2;

@Log4j2
public class World {

	public static final MyPoint GRAVITY = new MyPoint(0, Setting.getInstance().getGravity());
	private final double width;

	private final double height;

	private Object entitiesLock = new Object();

	private List<DrawableSimulable> entities;
	private Collection<DrawableSimulable> entitiesToRemove = new LinkedList<>();
	private Collection<DrawableSimulable> entitiesToAdd = new LinkedList<>();
	private List<UfoDestroyLog> destroyLogs = new LinkedList<>();

	private boolean viewMode;

	private Cannon cannon;

	public World(double width, double height) {
		this.width = width;
		this.height = height;
		entities = new ArrayList<>();
		cannon = new Cannon(this, new MyPoint(0, height - 20), -45);
		entities.add(cannon);
		entities.add(new Bullet(this, new MyPoint(0, height), new MyPoint(Setting.getInstance().getNormalBulletSpeed(),
				-Setting.getInstance().getNormalBulletSpeed()), GRAVITY));
		for (int i = 0; i < Setting.getInstance().getNumberOfUfos(); i++) {
			entities.add(new Ufo(this));
		}
	}

	public void setViewMode(boolean viewMode) {
		this.viewMode = viewMode;
		if (!viewMode) {
			startServer();
		} else {
			connectToServer();
		}
	}

	public void startServer() {
		new Thread(() -> {
			try (ServerSocket listenSocket = new ServerSocket(4600)) {
				while (!Thread.currentThread().isInterrupted()) {
					try {
						Socket client = listenSocket.accept();
						new Thread(() -> handleViewer(client), "Viewer handler").start();
					} catch (IOException e) {
						e.printStackTrace();
					}
				}
			} catch (IOException e) {
				e.printStackTrace();
			}

		}).start();
	}

	public void handleViewer(Socket client) {
		try (OutputStream outputStream = client.getOutputStream();
				ObjectOutputStream objectOutputStream = new ObjectOutputStream(outputStream)) {
			while (!Thread.currentThread().isInterrupted()) {
				synchronized (entitiesLock) {
					objectOutputStream.reset();
					objectOutputStream.writeObject(entities);
					objectOutputStream.flush();
				}
				sleepForWhile();
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void sleepForWhile() {
		try {
			Thread.sleep(20);
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
		}
	}

	public void connectToServer() {
		new Thread(() -> {
			try (Socket socketToServer = new Socket("localhost", 4600);
					InputStream inputStream = socketToServer.getInputStream();
					ObjectInputStream objectInputStream = new ObjectInputStream(inputStream)) {
				while (!Thread.currentThread().isInterrupted()) {
					Object o = objectInputStream.readObject();
					if (o instanceof List l) {
						synchronized (entitiesLock) {
							entities = l;
						}
					}
				}
			} catch (IOException | ClassNotFoundException e) {
				e.printStackTrace();
			}

		}).start();
	}

	public void add(UfoDestroyLog ufoDestroyLog) {
		destroyLogs.add(ufoDestroyLog);
	}

	public void pringDestroylog() {
		for (UfoDestroyLog ufoDestroyLog : destroyLogs) {
			log.info(ufoDestroyLog);
		}
	}

	public void draw(GraphicsContext gc) {
		gc.clearRect(0, 0, width, height);

		gc.save();
		synchronized (entitiesLock) {
			for (DrawableSimulable entity : entities) {
				entity.draw(gc);
			}
		}
		gc.restore();
	}

	public void simulate(double deltaT) {
		if (viewMode) {
			return;
		}
		synchronized (entitiesLock) {
			for (DrawableSimulable entity : entities) {
				entity.simulate(deltaT);
			}
			detectCollision();
			entities.removeAll(entitiesToRemove);
			entities.addAll(entitiesToAdd);
			entitiesToAdd.clear();
			entitiesToRemove.clear();
		}
	}

	private void detectCollision() {
		for (int i = 0; i < entities.size(); i++) {
			if (entities.get(i) instanceof Collisionable c1) {
				for (int j = i + 1; j < entities.size(); j++) {
					if (entities.get(j) instanceof Collisionable c2 && c1.intersect(c2.getBoundingBox())) {
						c1.hitBy(c2);
						c2.hitBy(c1);
					}
				}
			}
		}
	}

	public double getWidth() {
		return width;
	}

	public void add(DrawableSimulable entity) {
		if (viewMode) {
			return;
		}
		entitiesToAdd.add(entity);
	}

	public void remove(DrawableSimulable entity) {
		if (viewMode) {
			return;
		}
		entitiesToRemove.add(entity);

	}

	public double getHeight() {
		return height;
	}

	public Cannon getCannon() {
		return cannon;
	}
}